@@ -1,9 +1,9 @@
 ;;; eieio-opt.el -- eieio optional functions (debug, printing, speedbar)
 
-;;; Copyright (C) 1996, 1998, 1999, 2000, 2001 Eric M. Ludlam
+;;; Copyright (C) 1996, 1998, 1999, 2000, 2001, 2002 Eric M. Ludlam
 ;;
 ;; Author: <zappo@gnu.org>
-;; RCS: $Id: eieio-opt.el,v 1.17 2001/02/19 02:29:39 zappo Exp $
+;; RCS: $Id: eieio-opt.el,v 1.4 2002/11/28 17:30:38 scop Exp $
 ;; Keywords: OO, lisp
 ;;                                                                          
 ;; This program is free software; you can redistribute it and/or modify
@@ -81,6 +81,8 @@ Argument CH-PREFIX is another character prefix to display."
 If CLASS is actually an object, then also display current values of that obect."
   (interactive (list (eieio-read-class "Class: ")))
   (with-output-to-temp-buffer "*Help*"
+    (if (class-option class :abstract)
+	(princ "Abstract "))
     (princ "Class ")
     (prin1 class)
     (terpri)
@@ -215,27 +217,44 @@ Outputs to the standard output."
 	    prot (cdr prot)
 	    i (1+ i)))))
 
-(defun eieio-build-class-alist (&optional class buildlist)
+(defun eieio-build-class-alist (&optional class instantiable-only buildlist)
   "Return an alist of all currently active classes for completion purposes.
 Optional argument CLASS is the class to start with.
-Optional argument BUILDLIST is more list to attach."
+If INSTANTIABLE-ONLY is non nil, only allow names of classes which
+are not abstract, otherwise allow all classes.
+Optional argument BUILDLIST is more list to attach and is used internally."
   (let* ((cc (or class eieio-default-superclass))
 	 (sublst (aref (class-v cc) class-children)))
-    (setq buildlist (cons (cons (symbol-name cc) 1) buildlist))
+    (if (or (not instantiable-only) (not (class-abstract-p cc)))
+	(setq buildlist (cons (cons (symbol-name cc) 1) buildlist)))
     (while sublst
-      (setq buildlist (eieio-build-class-alist (car sublst) buildlist))
+      (setq buildlist (eieio-build-class-alist
+		       (car sublst) instantiable-only buildlist))
       (setq sublst (cdr sublst)))
     buildlist))
 
 (defvar eieio-read-class nil
   "History of the function `eieio-read-class' prompt.")
 
-(defun eieio-read-class (prompt &optional histvar)
+(defun eieio-read-class (prompt &optional histvar instantiable-only)
   "Return a class chosen by the user using PROMPT.
-Optional argument HISTVAR is a variable to use as history."
+Optional argument HISTVAR is a variable to use as history.
+If INSTANTIABLE-ONLY is non nil, only allow names of classes which
+are not abstract."
   (intern (completing-read prompt (eieio-build-class-alist) nil t nil
 			   (or histvar 'eieio-read-class))))
 
+(defun eieio-read-subclass (prompt class &optional histvar instantiable-only)
+  "Return a class chosen by the user using PROMPT.
+CLASS is the base class, and completion occurs across all subclasses.
+Optional argument HISTVAR is a variable to use as history.
+If INSTANTIABLE-ONLY is non nil, only allow names of classes which
+are not abstract."
+  (intern (completing-read prompt
+			   (eieio-build-class-alist class instantiable-only)
+			   nil t nil
+			   (or histvar 'eieio-read-class))))
+
 ;;; Collect all the generic functions created so far, and do cool stuff.
 ;;
 ;;;###autoload
@@ -379,16 +398,13 @@ Optional argument HISTORYVAR is the variable to use as history."
 
 ;;; Help system augmentation
 ;;
-(defun eieio-help-mode-augmentation-maybee ()
+(defun eieio-help-mode-augmentation-maybee (&optional ignored)
   "For buffers thrown into help mode, augment for eieio."
   ;; Scan created buttons so far if we are in help mode.
   (when (eq major-mode 'help-mode)
-    ;; View mode's read-only status of existing *Help* buffer is lost
-    ;; by with-output-to-temp-buffer.
-    (toggle-read-only -1)
-    (goto-char (point-min))
     (save-excursion
-      (let ((pos t))
+      (goto-char (point-min))
+      (let ((pos t) (inhibit-read-only t))
 	(while pos
 	  (if (get-text-property (point) 'help-xref) ; move off reference
 	      (goto-char
@@ -478,24 +494,33 @@ Optional argument HISTORYVAR is the variable to use as history."
   "Create buttons in speedbar that represents the current project.
 DIR-OR-OBJECT is the object to expand, or nil, and DEPTH is the current
 expansion depth."
-  ;; This function is only called once, to start the whole deal.
-  ;; Ceate, and expand the default object.
-  (eieio-class-button eieio-default-superclass 0)
-  (forward-line -1)
-  (speedbar-expand-line))
+  (when (eq (point-min) (point-max))
+    ;; This function is only called once, to start the whole deal.
+    ;; Ceate, and expand the default object.
+    (eieio-class-button eieio-default-superclass 0)
+    (forward-line -1)
+    (speedbar-expand-line)))
 
 (defun eieio-class-button (class depth)
   "Draw a speedbar button at the current point for CLASS at DEPTH."
   (if (not (class-p class))
       (signal 'wrong-type-argument (list 'class-p class)))
-  (speedbar-make-tag-line 'angle ?+
-			  'eieio-sb-expand
-			  class
-			  (symbol-name class)
-			  'eieio-describe-class-sb
-			  class
-			  'speedbar-directory-face
-			  depth))
+  (let ((subclasses (aref (class-v class) class-children)))
+    (if subclasses
+	(speedbar-make-tag-line 'angle ?+
+				'eieio-sb-expand
+				class
+				(symbol-name class)
+				'eieio-describe-class-sb
+				class
+				'speedbar-directory-face
+				depth)
+      (speedbar-make-tag-line 'angle ? nil nil
+			      (symbol-name class)
+			      'eieio-describe-class-sb
+			      class
+			      'speedbar-directory-face
+			      depth))))
 
 (defun eieio-sb-expand (text class indent)
   "For button TEXT, expand CLASS at the current location.
